unit Voxel;


interface

const TRANSPARENT = 256;

type

EError = (OK, ReadFailed, WriteFailed, InvalidSpanDataSizeCalced, InvalidSpan,
     BadSpan_SecondVoxelCount, Unhandled_Exception);

TVoxelHeader = packed record
     FileType: packed array[1..16] of Char; // always "Voxel Animation"
     Unknown, // always 1
     NumSections,
     NumSections2,
     BodySize: LongInt; // as above
     StartPaletteRemap,
     EndPaletteRemap: Byte; // colour indexes
     PaletteData: packed array[1..256*3] of Byte; // never used
end;

TVoxelSectionHeader = packed record
     Name: array[1..16] of Char; // ASCIIZ
     Number, // ordinal
     Unknown1, // always 1
     Unknown2: LongInt; // always 2
end;

TVoxelSectionTailer = packed record
     SpanStartOfs,
     SpanEndOfs,
     SpanDataOfs: LongInt;
     Det: Single;
     Transform: packed array[1..3,1..4] of Single;
     MinBounds, MaxBounds: packed array[1..3] of Single;
     XSize,
     YSize,
     ZSize,
     Unknown: Byte; // always 2 (or 4?); possibly normals-encoding scheme selection
end;

TVoxelUnpacked = record
     Colour,
     Normal,
     Flags: Byte;
     Used: Boolean;
end;

TVoxelPacked = LongInt;

TVoxelSection = class
     private
        // for storing / accessing the stored data
        Data: array of array of array of TVoxelPacked; // as is 32-bit type, should be packed anyway
        procedure SetDataSize(XSize,YSize,ZSize: Integer);
        procedure DefaultTransforms;
        function PackVoxel(Unpacked: TVoxelUnpacked): TVoxelPacked;
        procedure UnpackVoxel(PackedVoxel: TVoxelPacked; var dest: TVoxelUnpacked);
     public
        MaxNormal, // the highest normal value
        X, Y, Z: Integer; // cursor location
        // other
        Header: TVoxelSectionHeader;
        Tailer: TVoxelSectionTailer;
        constructor Create(); overload;
        constructor Create(Name: string; Number, XSize,YSize,ZSize: Integer); overload;
        procedure Resize(XSize,YSize,ZSize: Integer);

        procedure Clear; // blanks the entire voxel model
        procedure SetVoxel(x,y,z: Integer; src: TVoxelUnpacked);
        procedure GetVoxel(x,y,z: Integer; var dest: TVoxelUnpacked);
        // loading and saving
        function LoadFromFile(var F: File; HeadOfs, BodyOfs, TailOfs : Integer): EError;
//      function SaveToFileHeader(var F: File): EError;
        function SaveToFileBody(var F: File): EError;
//      function SaveToFileTailer(var F: File): EError;
        // utility methods
        function Name: string;
        // new by Koen
        procedure SetHeaderName(Name: String);

end;
//PVoxelSection = ^TVoxelSection;

TVoxel = class
     private
        Loaded: Boolean;
     public
        ErrorCode: EError; // if an error, says what it is
        ErrorMsg, // if an error, says what it is
        Filename: string;
        Header: TVoxelHeader;
        Section: array of TVoxelSection;
        constructor Create;
        procedure LoadFromFile(Fname: string); // examine ErrorCode for success
        procedure SaveToFile(Fname: string); // examine ErrorCode for success
        function isOpen: Boolean;
end;

function AsciizToStr(var src: array of Char; maxlen: Byte): string;

implementation

uses SysUtils, Math, Dialogs;

function AsciizToStr(var src: array of Char; maxlen: Byte): string;
var ch: Char;
    i: Integer;
begin
     SetLength(Result,maxlen);
     for i := 1 to maxlen do begin
         ch := src[i-1];
         if ch = #0 then begin
           SetLength(Result,i-1);
           Break;
         end;
         Result[i] := ch;
     end;
end;

procedure DebugMsg(Msg: string; Typ: TMsgDlgType);
begin
     // in this version, just show a messsage box for the user
     if (MessageDlg('Debug:' + Chr(10) + Msg,Typ,[mbOK,mbCancel],0) > 0) then
        Halt;
end;


(*** TVoxelSection members *********************************************)

procedure TVoxelSection.DefaultTransforms;
var i, j: integer;
begin
     Tailer.Det := 0.83333; // TODO: review if this is correct in all cases etc
     // TODO: check this logic is correct- was done from memory!
     for i := 1 to 3 do
         for j := 1 to 3 do // from memory, don't think this transform is ever used.
             if i = j then
                Tailer.Transform[i,j] := 1
             else
                 Tailer.Transform[i,j] := 0;
     // TODO: check this logic is correct- was done from memory!
     Tailer.MaxBounds[1] := (Tailer.XSize / 2) * Tailer.Det;
     Tailer.MaxBounds[2] := (Tailer.YSize / 2) * Tailer.Det;
     Tailer.MaxBounds[3] := (Tailer.ZSize / 2) * Tailer.Det;
     Tailer.MinBounds[1] := 0 - Tailer.MaxBounds[1];
     Tailer.MinBounds[2] := 0 - Tailer.MaxBounds[2];
     Tailer.MinBounds[3] := 0 - Tailer.MaxBounds[3];
end;

procedure TVoxelSection.Resize(XSize,YSize,ZSize: Integer);
var i: integer;
begin
     // memory alloc
     SetDataSize(XSize,YSize,ZSize); // will preserve contents where possible I believe
     // set tailer
     Tailer.XSize := XSize;
     Tailer.YSize := YSize;
     Tailer.ZSize := ZSize;
     DefaultTransforms;
end;

procedure TVoxelSection.SetHeaderName(Name: String);
  const
    MAX_LEN = 16;
  var
    i: integer;
  begin
    for i:=1 to 16 do
      Header.Name[i]:=#0;
    for i := 1 to Length(Name) do begin
      if i > MAX_LEN then break;
      Header.Name[i] := Name[i];
    end;
  end;

constructor TVoxelSection.Create(Name: string; Number, XSize,YSize,ZSize: Integer);
begin
     // allocate the memory
     SetDataSize(XSize,YSize,ZSize);
     // create header
     SetHeaderName(Name);
     Header.Number := Number;
     Header.Unknown1 := 1; // TODO: review if this is correct in all cases etc
     Header.Unknown2 := 2; // TODO: review if this is correct in all cases etc
     // create tailer
     Tailer.XSize := XSize;
     Tailer.YSize := YSize;
     Tailer.ZSize := ZSize;
     Tailer.Unknown := 2; // or 4 in RA2?  TODO: review if this is correct in all cases etc

     Tailer.Det:=1/12; //oops, forgot to set Det part correctly

     DefaultTransforms;
     // clear it
     Clear;
end;

constructor TVoxelSection.Create; begin end;

function TVoxelSection.Name: string;
begin
     Result := AsciizToStr(Header.Name,16);
end;

function TVoxelSection.PackVoxel(Unpacked: TVoxelUnpacked): TVoxelPacked;
begin
     Result := Unpacked.Flags shl 16;
     if Unpacked.Used then
        Result := Result or $100 or Unpacked.Colour
     else
         Result := Result or Unpacked.Colour;
     Result := Result shl 8;
     Result := Result or Unpacked.Normal;
end;

procedure TVoxelSection.UnpackVoxel(PackedVoxel: TVoxelPacked; var dest: TVoxelUnpacked);
begin
     dest.Normal := (PackedVoxel and $000000FF);
     dest.Colour := (PackedVoxel and $0000FF00) shr 8;
     dest.Used :=   (PackedVoxel and $00010000) > 0;
     dest.Flags :=  (PackedVoxel and $FF000000) shr 24;
end;

procedure TVoxelSection.Clear; // blanks the entire voxel model
var x, y, z: Integer;
    Empty: TVoxelUnpacked;
    PackedVoxel: TVoxelPacked;
begin
// prepare empty voxel
   with Empty do begin
        Colour := 0;
        Normal := 0;
        Used := False;
   end;
   PackedVoxel := PackVoxel(Empty);
// and set each and every voxel
     for x := 0 to (Tailer.XSize - 1) do
         for y := 0 to (Tailer.YSize - 1) do
             for z := 0 to (Tailer.ZSize - 1) do
                 Data[x,y,z] := PackedVoxel;
end;

procedure TVoxelSection.SetDataSize(XSize,YSize,ZSize: Integer);
var x, y: Integer;
begin
     SetLength(Data,XSize);
     for x := 0 to (XSize - 1) do begin
         SetLength(Data[x],YSize);
         for y := 0 to (YSize - 1) do
             SetLength(Data[x,y],ZSize);
     end;
end;

//This catches the crashes!
{$RANGECHECKS ON}
procedure TVoxelSection.SetVoxel(x,y,z: Integer; src: TVoxelUnpacked);
begin
     Data[x,y,z] := PackVoxel(src);
end;

procedure TVoxelSection.GetVoxel(x,y,z: Integer; var dest: TVoxelUnpacked);
begin
     UnpackVoxel(Data[x,y,z],dest);
end;

{$OPTIMIZATION OFF}
{$RANGECHECKS ON}
function TVoxelSection.LoadFromFile(var F: File; HeadOfs, BodyOfs, TailOfs : Integer): EError;
var BytesToRead,
    BytesRead,
    i, SpanCount, Ofs,
    x, y, z,
    voxel_num, num_voxels,
    SpanDataLen: Integer;
    UnpackedVoxel: TVoxelUnpacked;
    SpanStart, SpanEnd: packed array of LongInt;
    SpanData: packed array of Byte; // the raw data
procedure CleanUp(Err: Boolean);
begin
     SetLength(SpanData,0);
     SetLength(SpanEnd,0);
     SetLength(SpanStart,0);
     if Err then
        SetDataSize(0,0,0);
end;
begin
  MaxNormal := 0;
  try
     Result := OK; // assume ok
// read in tailer first
     Seek(F,TailOfs); // move to tail offset
     BytesToRead := SizeOf(TVoxelSectionTailer);
     BlockRead(F,Tailer,BytesToRead,BytesRead);
     if (BytesToRead <> BytesRead) then begin
        Result := ReadFailed;
        CleanUp(true);
        Exit;
     end;
     {with Tailer do
          DebugMsg('Size = [' + IntToStr(XSize) + ',' + IntToStr(YSize) + ',' + IntToStr(ZSize) + ']',mtInformation);}
// read in head next
     Seek(F,HeadOfs);
     BytesToRead := SizeOf(TVoxelSectionHeader);
     BlockRead(F,Header,BytesToRead,BytesRead);
     if (BytesToRead <> BytesRead) then begin
        Result := ReadFailed;
        CleanUp(True);
        Exit;
     end;
{$IFDEF DEBUG_NORMALS}
     DebugMsg('Header.Section name is : ' + Chr(10) + '[' + AsciizToStr(Header.Name,16) + ']' + Chr(10) +
        'Header.Number is : ' + IntToStr(Header.Number) + Chr(10) +
        'Header.Unknown1 is : ' + IntToStr(Header.Unknown1) + Chr(10) +
        'Header.Unknown2 is : ' + IntToStr(Header.Unknown2) + Chr(10) +
        'Tailer.Unknown is : ' + IntToStr(Tailer.Unknown)
        ,mtInformation);
{$ENDIF}
// create memory for storing the actual voxels; note the creation order
   with Tailer do begin
        SetDataSize(XSize,YSize,ZSize);
        Clear; // empty data
        SpanCount := XSize * YSize;
   end;
   SetLength(SpanStart,SpanCount);
   SetLength(SpanEnd,SpanCount);
// read in the span start offsets
   Seek(F,BodyOfs + Tailer.SpanStartOfs);
   BytesToRead := SpanCount * SizeOf(LongInt);
   BlockRead(F,SpanStart[0],BytesToRead,BytesRead);
   if (BytesToRead <> BytesRead) then begin
      Result := ReadFailed;
      CleanUp(True);
      Exit;
   end;
   Seek(F,BodyOfs + Tailer.SpanEndOfs);
   BlockRead(F,SpanEnd[0],BytesToRead,BytesRead);
   if (BytesToRead <> BytesRead) then begin
      Result := ReadFailed;
      CleanUp(True);
      Exit;
   end;
// and load the spans data into a buffer to be parsed
   // find last span end
   i := SpanCount - 1;
   while (SpanEnd[i] = -1) and (i > 0) do
         Dec(i);
   SpanDataLen := SpanEnd[i];
   // find first span end
   i := 0;
   while (SpanStart[i] = -1) and (i < SpanCount) do
         Inc(i);
   Dec(SpanDataLen,SpanStart[i]);
   Inc(SpanDataLen); // safety
   if SpanDataLen < 1 then begin
      Result := InvalidSpanDataSizeCalced;
      CleanUp(True);
      Exit;
   end;
   SetLength(SpanData,SpanDataLen);
   BlockRead(F,SpanData[0],SpanDataLen,BytesRead);
   if (SpanDataLen <> BytesRead) then begin
      Result := ReadFailed;
      CleanUp(True);
      Exit;
   end;
// and parse the spans
   i := -1; // will be inc'ed before anything else
   UnpackedVoxel.Used := True;
   for y := 0 to (Tailer.YSize - 1) do begin
       for x := 0 to (Tailer.XSize - 1) do begin
           Inc(i);
           if (SpanStart[i] = -1) or (SpanEnd[i] = -1) then // skip empty span?
              Continue;
           z := 0;
           Ofs := SpanStart[i];
           while (z < Tailer.ZSize) and (z < SpanEnd[i]) do begin
                 // this is one z - span
                 Inc(z,SpanData[Ofs]); Inc(Ofs); // skip count
                 num_voxels := SpanData[Ofs]; Inc(Ofs);
                 if (z >= Tailer.ZSize) then
                    Break;
                 if (z + num_voxels > Tailer.ZSize) then
                    DebugMsg('Err',mtError);
                 for voxel_num := 1 to num_voxels do begin
                     with UnpackedVoxel do begin
                          Colour := SpanData[Ofs]; Inc(Ofs);
                          Normal := SpanData[Ofs]; Inc(Ofs);
                          if (Normal > MaxNormal) and (Normal <> $FF) then
                             MaxNormal := Normal;
                     end;
                     SetVoxel(x,y,z,UnpackedVoxel);
                     Inc(z);
                 end;
                 // check second span number
                 voxel_num := SpanData[Ofs]; Inc(Ofs);
                 if voxel_num <> num_voxels then begin
                    Result := BadSpan_SecondVoxelCount;
                    CleanUp(True);
                    Exit;
                 end;
           end;
       end;
   end;
{$IFDEF DEBUG_NORMALS}
   DebugMsg('Max normals = ' + IntToStr(MaxNormal),mtInformation);
{$ENDIF}
// done
  except
        Result := Unhandled_Exception;
        CleanUp(True);
  end;
  // done
  CleanUp(False);
end;

function TVoxelSection.SaveToFileBody(var F: File): EError;
const MaskUsed: TVoxelPacked = $00010000;
function SpanUsed(x, y: integer): boolean;
var z: integer;
begin
     Result := True; // assume it is
     for z := 0 to (Tailer.ZSize - 1) do
         if (Data[x,y,z] and MaskUsed) > 0 then // used flag set?
            Exit;
     Result := False; // if here, span wasn't used after all
end;
function SpanLength(x,y,z: integer): integer;
var v: TVoxelPacked;
begin
     Result := 0;
     while z < Tailer.ZSize do begin
           v := Data[x,y,z];
           if (v and MaskUsed) > 0 then
              Inc(Result)
           else
               Exit;
           Inc(z);
     end;
end;
var FSpanStart, FSpanEnd, FBody,
    SpanStart, SpanEnd: LongInt;
    x, y, z, SpanSize, s: integer;
    v: TVoxelPacked;
    skip, spanlen, colour, normal: byte;
begin
     // work out offsets
     FBody := FilePos(F);
     SpanSize := Tailer.XSize * Tailer.YSize * SizeOf(LongInt);
     FSpanEnd := FBody - SpanSize;
     FSpanStart := FSpanEnd - SpanSize;
     // and save some info now to the tailer
     Tailer.SpanStartOfs := FSpanStart;
     Tailer.SpanEndOfs := FSpanEnd;
     Tailer.SpanDataOfs := FBody;
     for y := 0 to (Tailer.YSize - 1) do
         for x := 0 to (Tailer.XSize - 1) do begin
             if not SpanUsed(x,y) then begin
                SpanStart := -1;
                SpanEnd := -1;
             end else begin
                 SpanStart := FilePos(F) - Tailer.SpanDataOfs;
                 skip := 0;
                 z := 0;
                 while z < (Tailer.ZSize) do begin
                     spanlen := SpanLength(x,y,z);
                     if (spanlen > 0) then begin
                        BlockWrite(F,skip,1); // write skip
                        BlockWrite(F,spanlen,1); // write span length again
                        for s := 1 to spanlen do begin
                            v := Data[x,y,z + s - 1]; // get voxel
                            colour := (v and $0000FF00) shr 8;
                            normal := (v and $000000FF);
                            BlockWrite(F,colour,1);
                            BlockWrite(F,normal,1);
                        end;
                        BlockWrite(F,spanlen,1); // write span length again
                        Inc(z,spanlen);
                        skip := 0;
                     end else begin
                         Inc(skip);
                         Inc(z);
                     end;
                 end;
                 if skip > 0 then begin // write a dummy?
                    BlockWrite(F,skip,1);
                    spanlen := 0;
                    BlockWrite(F,spanlen,1);
                    BlockWrite(F,spanlen,1);
                 end;
                 SpanEnd := FilePos(F) - Tailer.SpanDataOfs - 1;
             end;
             // write span data
             FBody := FilePos(F);
             Seek(F,FSpanStart);
             BlockWrite(F,SpanStart,SizeOf(SpanStart));
             Inc(FSpanStart,SizeOf(SpanStart));
             Seek(F,FSpanEnd);
             BlockWrite(F,SpanEnd,SizeOf(SpanEnd));
             Inc(FSpanEnd,SizeOf(SpanEnd));
             Seek(F,FBody); //return to span data
         end;
     Result := OK;
end;

{$OPTIMIZATION ON}
{$RANGECHECKS OFF}

constructor TVoxel.Create;
begin
     ErrorCode := OK;
     Filename := '';
     Loaded := False;
end;

procedure TVoxel.LoadFromFile(FName: string);
var F: File;
    BytesToRead,
    BytesRead,
    i,
    HeadOfs, TailOfs, BodyOfs: Integer;
begin
     {DebugMsg(      'SizeOf(Header) = ' + IntToStr(SizeOf(Header)) + Chr(10) +
        'SizeOf(TVoxelSectionHeader) = ' + IntToStr(SizeOf(TVoxelSectionHeader)) + Chr(10) +
        'SizeOf(TVoxelSectionTailer) = ' + IntToStr(SizeOf(TVoxelSectionTailer)),mtInformation);}
     ErrorCode := OK;
     Loaded := False;
     Filename := FName;
// open voxel file to read
     AssignFile(F,Filename);
     FileMode := 0; // read only
     Reset(F,1); // file of byte
// read in main header
     BytesToRead := SizeOf(Header);
     BlockRead(F,Header,BytesToRead,BytesRead);
     if (BytesRead <> BytesToRead) then begin
        ErrorCode := ReadFailed;
       // ErrorMsg := 'Could not read Voxel Header';
        //DebugMsg(ErrorMsg,mtError);
        CloseFile(F);
        Exit;
     end;
     //DebugMsg('there are ' + IntToStr(Header.NumSections) + ' sections',mtInformation);
// read in sections
     // prepare section objects
     SetLength(Section,Header.NumSections);
     for i := 1 to Header.NumSections do
         Section[i-1] := TVoxelSection.Create;
     HeadOfs := SizeOf(Header); // starts immediately after header
     BodyOfs := HeadOfs + (Header.NumSections * SizeOf(TVoxelSectionHeader)); // after all section headers
     TailOfs := BodyOfs + Header.BodySize; // last part of file
     // and load them
     for i := 1 to Header.NumSections do begin
         ErrorCode := Section[i-1].LoadFromFile(F,HeadOfs,BodyOfs,TailOfs);
         if (ErrorCode <> OK) then begin
            ErrorMsg := 'Could not read Voxel Section Tailer[' + IntToStr(i) + ']';
            DebugMsg(ErrorMsg,mtError);
            Exit;
         end;
         Inc(HeadOfs,SizeOf(TVoxelSectionHeader)); // next header
         // BodyOfs doesn't change, as relative offset is in tailer
         Inc(TailOfs,SizeOf(TVoxelSectionTailer)); // next tailer
  end;
  Loaded := True;
  CloseFile(F);
end;

procedure TVoxel.SaveToFile(Fname: string);
{ First, we create the file.  We write blanks to represent the
  header. We then write the section headers.  We then encode the
  section bodies.  We then return to the section headers and
  write them, then the section tailers.  We then write the voxel
  header.
}
var F: File;
  BytesToWrite,
  BytesWritten: integer;

  function WriteBlank(count: integer): EError;
  var
    i: integer;
    ch: byte;
  begin
    ch := 0;
    BytesToWrite := 1;
    for i := 1 to count do begin
      BlockWrite(F,ch,BytesToWrite,BytesWritten);
      if (BytesWritten <> BytesToWrite) then begin
        Result := WriteFailed;
        Exit;
      end;
    end;
    Result := OK;
  end;
var
  i, BodyStart, BodyEnd: Integer;
begin
     if not Loaded then Exit; // cannot save an empty voxel?
     //-- when we save an empty voxel - this voxel will be unsupported - we must change this [Kamil ^aka Plasmadroid]
     ErrorCode := OK;
     Filename := Fname;
// create voxel file to write
     AssignFile(F,Filename);
     Rewrite(F,1); // file of byte
// write main header (overwritten later)
     ErrorCode := WriteBlank(SizeOf(TVoxelHeader));
     if (ErrorCode <> OK) then begin
        ErrorMsg := 'Could not write Voxel Header (pass 2)';
        DebugMsg(ErrorMsg,mtError);
        Exit;
     end;
// write section heads (real data)
     BytesToWrite := SizeOf(TVoxelSectionHeader);
     for i := 1 to Header.NumSections do begin
         BlockWrite(F,Section[Pred(i)].Header,BytesToWrite,BytesWritten);
         if (BytesWritten <> BytesToWrite) then begin
            ErrorCode := WriteFailed;
            ErrorMsg := 'Could not write Voxel Section Header #' + IntToStr(i) + '(pass 2)';
            DebugMsg(ErrorMsg,mtError);
            Exit;
         end;
     end;
// write section bodies (real data)
     BodyStart := FilePos(F); // for later; could have calc'ed it instead
     for i := 0 to (Header.NumSections - 1) do begin
         with Section[i] do begin
              // first, write blank data where the span starts / ends will be
              ErrorCode := WriteBlank(Tailer.XSize*Tailer.YSize*4*2);
              if (ErrorCode <> OK) then begin
                 ErrorMsg := 'Could not write Voxel Section Body #' + IntToStr(Succ(i)) + ' (pass 1)';
                 DebugMsg(ErrorMsg,mtError);
                 Exit;
              end;
              // now write the actual data
              ErrorCode := SaveToFileBody(F);
              if (ErrorCode <> OK) then begin
                 ErrorMsg := 'Could not write Voxel Section Body #' + IntToStr(Succ(i)) + ' (pass 2)';
                 DebugMsg(ErrorMsg,mtError);
                 Exit;
              end;
         end;
     end;
     BodyEnd := FilePos(F); // for later; could have calc'ed it instead
// write section tailers (real data)
     BytesToWrite := SizeOf(TVoxelSectionTailer);
     for i := 0 to (Header.NumSections - 1) do begin
         with Section[i].Tailer do begin
              Dec(SpanStartOfs,BodyStart);
              Dec(SpanEndOfs,BodyStart);
              Dec(SpanDataOfs,BodyStart);
         end;
         BlockWrite(F,Section[i].Tailer,BytesToWrite,BytesWritten);
         if (BytesWritten <> BytesToWrite) then begin
            ErrorCode := WriteFailed;
            ErrorMsg := 'Could not write Voxel Section Tailer #' + IntToStr(i) + '(pass 2)';
            DebugMsg(ErrorMsg,mtError);
            Exit;
         end;
     end;
// write main header (real data)
     Seek(F,0); // go to start of file again
     Header.BodySize := BodyEnd - BodyStart;
     BytesToWrite := SizeOf(TVoxelHeader);
     BlockWrite(F,Header,BytesToWrite,BytesWritten);
     if (BytesWritten <> BytesToWrite) then begin
        ErrorMsg := 'Could not write Voxel Header (pass 2)';
        DebugMsg(ErrorMsg,mtError);
        Exit;
     end;
     CloseFile(F);
end;

function TVoxel.isOpen: Boolean; begin Result := Loaded; end;

end.
