unit Voxel_Tools;

interface

Uses Voxel;

// returns the number of redundant voxels removed
function RemoveRedundantVoxels(voxel: TVoxelSection): integer;

const FACING_LEFT = 2;
      FACING_RIGHT = FACING_LEFT * 2;
      FACING_TOP = FACING_RIGHT * 2;
      FACING_BOTTOM = FACING_TOP * 2;
      FACING_FRONT = FACING_BOTTOM * 2;
      FACING_BACK = FACING_FRONT * 2;
      FACING_MASK = FACING_LEFT or FACING_RIGHT or FACING_TOP or FACING_BOTTOM or FACING_FRONT or FACING_BACK;
      NOT_SET = -1;

// fills the Flags attribute of each voxel with facings information
// facings can be decoded by applying the above masks
// before running, always ensure that redundant voxels have been removed,
// else results will be screwed
procedure CalcFacings(voxel: TVoxelSection);

type TNormalsScheme = record
     author: string;
     scheme: integer; // NOT_SET means ignore this whole scheme because there was an error loading it
     normal: array[0..255] of integer; // NOT_SET means not used
end;
// loads a normals-scheme
function LoadNormalsScheme(filename: string): TNormalsScheme;
//applies a normals-scheme
procedure ApplyNormals(voxel: TVoxelSection; var NormalsScheme: TNormalsScheme);

// use these for basic text-based normals usage reports
type FacingReport = array[0..255,0..255] of integer;
function GenerateFacingReport(voxel: TVoxelSection): FacingReport;
function FacingUsed(facing: integer; var report: FacingReport): boolean;
procedure PrintFacingReport(var report: FacingReport; var outfile: Text);
procedure PrintFacing(facing: integer; var report: FacingReport; var outfile: Text);

function FacingAsStr(facing: integer): string;

implementation

uses SysUtils, IniFiles, Classes;

function BitSet(value, mask: integer): boolean;
begin
     Result := (value and mask) = mask;
end;

const isInside = 0;
      isOutside = 1;

procedure FindInside(voxel: TVoxelSection);
{ This function starts by assuming every cell is inside.
  It then does a flood fill from every face of the cube,
  flood-filling all empty cells to outside.
}
var maxx, maxy, maxz: integer;
procedure ClearFlags;
var v: TVoxelUnpacked;
    x, y, z: integer;
begin
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do begin
                 voxel.GetVoxel(x,y,z,v);
                 v.Flags := isInside;
                 voxel.SetVoxel(x,y,z,v);
             end;
end;
var OutsideCount: integer;
procedure SetOutside(x,y,z: integer); // recursive- like a flood-fill
var v: TVoxelUnpacked;
begin
     // bounds check!
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // get the voxel
     voxel.GetVoxel(x,y,z,v);
     // already scanned?
     if v.Flags = isOutside then Exit;
     // is solid? stops a flood-fill
     if v.Used then Exit;
     // set to outside
     v.Flags := isOutside;
     Inc(OutsideCount); // record count
     voxel.SetVoxel(x,y,z,v);
     // and recurse on each axis
     SetOutside(x - 1, y, z);
     SetOutside(x + 1, y, z);
     SetOutside(x, y - 1, z);
     SetOutside(x, y + 1, z);
     SetOutside(x, y, z - 1);
     SetOutside(x, y, z + 1);
end;
var x, y, z: integer;
begin
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     ClearFlags;
     OutsideCount := 0;
     // fill each X side
     for y := 0 to maxy do
         for z := 0 to maxz do begin
             SetOutside(0,y,z); // will not fill used voxels anyway
             SetOutside(maxx,y,z);
         end;
     // fill each Y side
     for x := 0 to maxx do
         for z := 0 to maxz do begin
             SetOutside(x,0,z);
             SetOutside(x,maxy,z);
         end;
     // fill each Z side
     for x := 0 to maxx do
         for y := 0 to maxy do begin
             SetOutside(x,y,0);
             SetOutside(x,y,maxz);
         end;
//   x := (maxx+1)*(maxy+1)*(maxz+1);
//   ShowMessage(IntToStr(OutsideCount) + ' cells are outside' + Chr(10) +
//   IntToStr(x) + ' cells in total');
end;

function RemoveRedundantVoxels(voxel: TVoxelSection): integer;
var maxx, maxy, maxz: integer;
function IsRedundant(x,y,z: integer): boolean;
function IsOnOutside(x,y,z: integer): boolean;
var v: TVoxelUnpacked;
begin
     Result := True; // assume it is
     // bounds check
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // get the voxel
     voxel.GetVoxel(x,y,z,v);
     // is outside?
     Result := (v.Flags = isOutside);
end;
begin
     Result := false; // assume it isn't
     if IsOnOutside(x,y,z) then Exit;
     // ok, examine neighbours
     if IsOnOutside(x - 1,y,z) then Exit;
     if IsOnOutside(x + 1,y,z) then Exit;
     if IsOnOutside(x,y - 1,z) then Exit;
     if IsOnOutside(x,y + 1,z) then Exit;
     if IsOnOutside(x,y,z - 1) then Exit;
     if IsOnOutside(x,y,z + 1) then Exit;
     // if here, must be surrounded by insides
     Result := true;
end;
var v: TVoxelUnpacked;
    x, y, z: integer;
begin
     Result := 0; // none removed yet
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // find the inside voxels
     FindInside(voxel);
     // examine each used voxel.  If it is beside all used inside voxels, then delete it
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 if IsRedundant(x,y,z) then begin
                    voxel.GetVoxel(x,y,z,v);
                    if v.Used then begin // ignore already empty cells
                       v.Used := false; // delete it!
                       voxel.SetVoxel(x,y,z,v);
                       Inc(Result); // one more removed
                    end;
                 end;
     // done!
end;

procedure CalcFacings(voxel: TVoxelSection);
var maxx, maxy, maxz,
    x, y, z: integer;
procedure CalcFacing;
function Empty(x,y,z: integer): boolean;
var v: TVoxelUnpacked;
begin
     // check bounds
     Result := True; // outside the voxel is all empty
     if (x < 0) or (x > maxx) then Exit;
     if (y < 0) or (y > maxy) then Exit;
     if (z < 0) or (z > maxz) then Exit;
     // ok, do the real check
     voxel.GetVoxel(x,y,z,v);
     if v.Used then
        Result := False
     else
         Result := BitSet(v.Flags,isOutside);
end;
var v: TVoxelUnpacked;
    Flags: byte;
begin
     // get the voxel in question
     voxel.GetVoxel(x,y,z,v);
     // skip empty ones
     if not v.Used then Exit;
     Flags := v.Flags;
     // facings
     if Empty(x-1,y,z) then Flags := (Flags or FACING_FRONT);
     if Empty(x+1,y,z) then Flags := (Flags or FACING_BACK);
     if Empty(x,y-1,z) then Flags := (Flags or FACING_LEFT);
     if Empty(x,y+1,z) then Flags := (Flags or FACING_RIGHT);
     if Empty(x,y,z-1) then Flags := (Flags or FACING_BOTTOM);
     if Empty(x,y,z+1) then Flags := (Flags or FACING_TOP);
     // cancel out double-sided facings
     // TODO: if x is less than half-way, prefer front, else back etc
     if BitSet(Flags,FACING_FRONT or FACING_BACK) then Dec(Flags,FACING_BACK);
     if BitSet(Flags,FACING_LEFT or FACING_RIGHT) then Dec(Flags,FACING_RIGHT);
     if BitSet(Flags,FACING_TOP or FACING_BOTTOM) then Dec(Flags,FACING_BOTTOM);
     // save the result
     v.Flags := Flags;
     voxel.SetVoxel(x,y,z,v);
end;
begin
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // examine each used voxel.
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 CalcFacing;
end;

function GenerateFacingReport(voxel: TVoxelSection): FacingReport;
procedure Clear;
var facing, normal: integer;
begin
     for facing := 0 to 255 do
         for normal := 0 to 255 do
             Result[facing,normal] := 0;
end;
var x, y, z,
    maxx, maxy, maxz: integer;
    v: TVoxelUnpacked;
begin
     Clear;
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // examine each used voxel.
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do begin
                 voxel.GetVoxel(x,y,z,v);
                 if v.Used then
                    Inc(Result[v.Flags and FACING_MASK,v.Normal]);
             end;
end;

function NormalsCount(facing: integer; var report: FacingReport): integer;
var normal: integer;
begin
     Result := 0;
     for normal := 0 to 255 do
         Inc(Result,report[facing,normal]);
end;

function FacingUsed(facing: integer; var report: FacingReport): boolean;
var normal, count: integer;
begin
     count := 0;
     for normal := 0 to 255 do
         Inc(count,report[facing,normal]);
     Result := (count > 0);
end;

procedure PrintFacing(facing: integer; var report: FacingReport; var outfile: Text);
var normal, sum, percent, astrix: integer;
begin
     sum := NormalsCount(facing,report);
     WriteLn(outfile,'facing ',facing shr 1,' ',FacingAsStr(facing),' (');
     for normal := 0 to 255 do
         if report[facing,normal] > 0 then begin // used?
            // data
            Write(outfile,#9,#9,normal:3,' = ',report[facing,normal]:3,'  ');
            // percent bar
            percent := Round((report[facing,normal] / sum) * 50); // 1 astrix per percent
            for astrix := 1 to percent do
                Write(outfile,'*');
            WriteLn(outfile);
         end;
     WriteLn(outfile,#9,')');
end;

procedure PrintFacingReport(var report: FacingReport; var outfile: Text);
var facing: integer;
begin
     for facing := 0 to 255 do begin
         if FacingUsed(facing,report) then
            PrintFacing(facing,report,outfile);
     end;
end;

function FacingAsStr(facing: integer): string;
begin
     Result := '';
     if BitSet(facing,FACING_LEFT) then Result := Concat(Result,' LEFT');
     if BitSet(facing,FACING_RIGHT) then Result := Concat(Result,' RIGHT');
     if BitSet(facing,FACING_TOP) then Result := Concat(Result,' TOP');
     if BitSet(facing,FACING_BOTTOM) then Result := Concat(Result,' BOTTOM');
     if BitSet(facing,FACING_FRONT) then Result := Concat(Result,' FRONT');
     if BitSet(facing,FACING_BACK) then Result := Concat(Result,' BACK');
     Result := Trim(Result);
     if Result = '' then Result := 'CONFUSED';
end;

// loads a normals-scheme
function LoadNormalsScheme(filename: string): TNormalsScheme;

var F: TStrings;
    Key, Value: String;
    Section: integer;

procedure Replace(var s: string; const oldchar, newchar: char);
var i: integer;
begin
     for i := 1 to Length(s) do
         if s[i] = oldchar then
            s[i] := newchar;
end;

function Split(const Line: integer; var Key: String; var Value: String): boolean; // returns false if a bad line
var equals: integer;
    s: String;
begin
     Result := False; // assume bad
     s := F[Line];
     Replace(s,#9,#32); // replace tabs with spaces
     s := Trim(s);
     equals := Pos('=',s);
     if equals <= 1 then Exit; // no separator or starts with a space?
     Key := Trim(Copy(s,1,Pred(equals)));
     Value := Trim(Copy(s,Succ(equals),100));
     Result := True; // ok
end;

function isSection(const Line: integer): boolean;
var s: String;
begin
     s := F[Line];
     Replace(s,#9,#32); // replace tabs with spaces
     s := Trim(s);
     Result := (Pos('[',s) = 1) and (Pos(']',s) = Length(s));
end;

function GetSectionName(const Line: integer): String;
var s: String;
begin
     s := F[Line];
     Replace(s,#9,#32); // replace tabs with spaces
     s := Trim(s);
     if (Pos('[',s) = 1) and (Pos(']',s) = Length(s)) then
        Result := Trim(Copy(s,2,Length(s)-2))
     else
         Result := '';
end;

function SetSection(const name: String): boolean; // return FALSE if section not found
var sec: integer;
begin
     for Sec := 0 to F.Count - 1 do begin
         Section := Sec;
         if UpperCase(GetSectionName(Sec)) = Trim(UpperCase(name)) then begin
            Result := True;
            Exit;
         end;
     end;
     Result := False;
end;

function GetAsString(const name: String; const def: String): String; // return default if key not found
var Line: integer;
begin
     Result := def;
     Line := Succ(Section);
     while Line < F.Count do begin
           if IsSection(Line) then
              Break;
           Split(Line,Key,Value);
           if UpperCase(Key) = UpperCase(Trim(name)) then begin
              Result := Value;
              Break;
           end;
           Inc(Line);
     end;
end;

function GetAsInt(const name: String; const def: integer): integer; // return default if key not found
var Line: integer;
begin
     Result := def;
     try
        Line := Succ(Section);
        while Line < F.Count do begin
           if IsSection(Line) then
              Break;
           Split(Line,Key,Value);
           if UpperCase(Key) = UpperCase(Trim(name)) then begin
              Result := StrToInt(Value);
              Break;
           end;
           Inc(Line);
        end;
     except
        on EConvertError do begin
           WriteLn('Warning: ',Key,' is not set to an integer value!');
           Result := def;
        end;
     end;
end;

procedure Clear;
var normal: integer;
begin
     Result.author := 'anonymous';
     Result.scheme := NOT_SET; // indicates bad
     for normal := 0 to 255 do
         Result.normal[normal] := NOT_SET; // indicates not assigned
end;
var Scheme, uKey: String;
    Facing: integer;
    Line, Normal: integer;
    E: Exception;
begin
     try
          WriteLn('Loading normals scheme from "',filename,'"..');
          Clear;
          if not FileExists(filename) then begin
             WriteLn('Error: file does not exist');
             Exit;
          end;
          // open file
          F := TStringList.Create;
          F.LoadFromFile(filename);
          // get details
          if not SetSection('details') then
             raise Exception.Create('details section not found');
          Result.author := GetAsString('author','anonymous');
          Scheme := GetAsString('scheme','');
          if UpperCase(Scheme) = 'TIBERIAN_SUN' then
             Result.scheme := 02
          else if UpperCase(Scheme) = 'RED_ALERT_2' then
               Result.scheme := 04
          else if Scheme = '' then
               raise Exception.Create('scheme not specified')
          else
              raise Exception.Create(Concat('scheme "',Scheme,'" not valid'));
          // load normals
          if not SetSection('normals') then
             raise Exception.Create('normals section not found');
          for Line := Succ(Section) to F.Count - 1 do begin
              if isSection(Line) then Break; // end of section?
              if not Split(Line,Key,Value) then
                 Continue; // skip empty lines
              // get value
              try
                 Normal := StrToInt(Value);
              except
                    on EConvertError do
                       Normal := NOT_SET;
              end;
              if Normal > NOT_SET then // out of range checks
                 if Result.scheme = 02 then begin // TS
                    if Normal > 35 then Normal := NOT_SET;
                 end else if Normal > 255 then // RA2
                     Normal := NOT_SET;
              if Normal <= NOT_SET then begin
                 WriteLn('Warning: The key "',Key,'" does not have a sensible value');
                 Normal := NOT_SET;
              end else begin
                  // determine mask
                  Facing := 0;
                  uKey := UpperCase(Key);
                  if Pos('LEFT',uKey) > 0 then Facing := Facing or FACING_LEFT;
                  if Pos('RIGHT',uKey) > 0 then Facing := Facing or FACING_RIGHT;
                  if Pos('TOP',uKey) > 0 then Facing := Facing or FACING_TOP;
                  if Pos('BOTTOM',uKey) > 0 then Facing := Facing or FACING_BOTTOM;
                  if Pos('FRONT',uKey) > 0 then Facing := Facing or FACING_FRONT;
                  if Pos('BACK',uKey) > 0 then Facing := Facing or FACING_BACK;
                  // no facings at all?
                  if Facing = 0 then
                     WriteLn('Warning: The key "',Key,'" has no valid facing(s)')
                  else begin
                       // crazy mask?
                       if (BitSet(Facing,FACING_LEFT) and BitSet(Facing,FACING_RIGHT)) or
                          (BitSet(Facing,FACING_TOP) and BitSet(Facing,FACING_BOTTOM)) or
                          (BitSet(Facing,FACING_FRONT) and BitSet(Facing,FACING_BACK)) then
                          WriteLn('Warning: The key "',Key,'" has an absurd facing(s)');
                       // duplicate?
                       if Result.Normal[Facing] <> NOT_SET then
                          WriteLn('Warning: The key "',Key,'" overwrites a previously specified value');
                       // save it
                       Result.normal[Facing] := Normal;
                  end;
              end;
          end;
          // done
          F.Free;
     except
           on E: Exception do begin
              WriteLn('Error: ',E.Message);
              Result.scheme := NOT_SET;
           end;
     end;
     WriteLn('done.');
end;

//applies a normals-scheme
procedure ApplyNormals(voxel: TVoxelSection; var NormalsScheme: TNormalsScheme);
var maxx, maxy, maxz,
    x, y, z: integer;
procedure ApplyNormal;
var v: TVoxelUnpacked;
    Facing, Normal: integer;
begin
     voxel.GetVoxel(x,y,z,v);
     if v.Used then begin
        Facing := v.Flags and FACING_MASK;
        Normal := NormalsScheme.normal[Facing];
        if Normal <> NOT_SET then begin
           v.Normal := Normal;
           voxel.SetVoxel(x,y,z,v);
        end else
            WriteLn('Voxel at (',x,',',y,',',z,') has unspecified facing (',FacingAsStr(Facing),')');
     end;
end;
begin
     // constraints
     maxx := voxel.Tailer.XSize - 1;
     maxy := voxel.Tailer.YSize - 1;
     maxz := voxel.Tailer.ZSize - 1;
     // set right mode
     voxel.Tailer.Unknown := NormalsScheme.scheme;
     // examine each used voxel.
     for x := 0 to maxx do
         for y := 0 to maxy do
             for z := 0 to maxz do
                 ApplyNormal;
end;

end.
